@@ -18,55 +18,12 @@ from StringIO import StringIO
 from sprinter import lib
 from sprinter.dependencytree import DependencyTree, DependencyTreeException
 
-test_old_version = """
-[config]
-namespace = sprinter
-
-[maven]
-formula = sprinter.formulas.unpack
-specific_version = 2.10
-
-[ant]
-formula = sprinter.formulas.unpack
-specific_version = 1.8.4
-
-[mysql]
-formula = sprinter.formulas.package
-apt-get = libmysqlclient
-          libmysqlclient-dev
-brew = mysql
-"""
-
-test_new_version = """
-[config]
-namespace = sprinter
-
-[maven]
-formula = sprinter.formulas.unpack
-specific_version = 3.0.4
-
-[ant]
-formula = sprinter.formulas.unpack
-specific_version = 1.8.4
-
-[myrc]
-formula = sprinter.formulas.template
-"""
-
-test_input_string = \
-"""
-gitroot==~/workspace
-username
-password?
-main_branch==comp_main
-"""
-
-
 CONFIG_RESERVED = ['source', 'inputs']
+FEATURE_RESERVED = ['rc', 'command', 'phase']
 NAMESPACE_REGEX = re.compile('([a-zA-Z0-9_]+)(\.[a-zA-Z0-9_]+)?$')
 
 
-class ManifestError(Exception):
+class ManifestException(Exception):
     pass
 
 
@@ -84,10 +41,11 @@ class Manifest(object):
     """
     dtree = None  # dependency tree object to ascertain order
     invalidations = []   # a list of the invalidation of the Manifest. Aggregated while parsing
+    additional_context_variables = {}  # a list of the additional context variables available
 
-    def __init__(self, raw_manifest, namespace=None, logger='sprinter',
+    def __init__(self, raw_manifest, namespace=None, logger=None,
                  username=None, password=None):
-        self.logger = logging.getLogger(logger)
+        self.logger = logger if logger else logging.getLogger('sprinter')
         self.manifest = self.__load_manifest(raw_manifest,
                                              username=username,
                                              password=password)
@@ -101,6 +59,51 @@ class Manifest(object):
         return self.get('config', 'source') if \
             self.has_option('config', 'source') else None
 
+    def formula_sections(self):
+        """
+        Return all sections related to a formula, re-ordered according to the "depends" section.
+        """
+        if self.dtree is not None:
+            return self.dtree.order
+        else:
+            return [s for s in self.manifest.sections() if s != "config"]
+
+    def is_true(self, section, option):
+        """
+        Return true if the section option combo exists and it is set
+        to a truthy value.
+        """
+        return self.has_option(section, option) and \
+            self.get(section, option).lower().startswith('t')
+
+    def get_feature_class(self, section):
+        if section not in self.formula_sections():
+            raise ManifestException("Cannot get feature %s!" % section)
+        return self.manifest.get(section, 'formula')
+
+    def get_feature_config(self, feature):
+        """ Get the feature configuration for a class """
+        context_dict = dict(self.additional_context_variables.items() +
+                            self.get_context_dict().items())
+        return self.__substitute_objects(dict(self.manifest.items(feature)), context_dict)
+
+    def get_context_dict(self):
+        """ return a context dict of the desired state """
+        context_dict = {}
+        for s in self.formula_sections():
+            for k, v in self.manifest.items(s):
+                context_dict["%s:%s" % (s, k)] = v
+        return context_dict
+
+    def run_phase(self, feature, phase_name):
+        """ Determine if the feature should run in the given phase """
+        return (not self.manifest.has_option(feature, 'phases') or
+                phase_name in [x.strip() for x in self.manifest.get(feature, 'phases').split(",")])
+
+    def add_additional_context(self, additonal_context):
+        """ Add additional context variable """
+        self.additional_context_variables = dict(self.additional_context_variables.items() + additonal_context.items())
+
     def __load_manifest(self, raw_manifest, username=None, password=None):
         manifest = RawConfigParser()
         manifest.add_section('config')
@@ -153,27 +156,43 @@ class Manifest(object):
             self.invalidations.append("Issue with building dependency tree: %s" % str(dte))
             return None
 
-    def formula_sections(self):
+    def __substitute_objects(self, value, context_dict):
         """
-        Return all sections related to a formula, re-ordered according to the "depends" section.
+        recursively substitute value with the context_dict
         """
-        if self.dtree is not None:
-            return self.dtree.order
+        if type(value) == dict:
+            return dict([(k, self.__substitute_objects(v, context_dict)) for k, v in value.items()])
+        elif type(value) == str:
+            try:
+                return value % context_dict
+            except KeyError as e:
+                self.logger.warn("Could not specialize %s! Error: %s" % (value, e))
+                return value
         else:
-            return [s for s in self.manifest.sections() if s != "config"]
+            return value
 
-    def valid(self):
-        """
-        Validate the configuration, ensure that the configuration is
-        properly formatted.
-        """
+    # custom equality method
+    def __eq__(self, other):
+        if not isinstance(other, Manifest):
+            return False
+        for s in self.manifest.sections():
+            if s not in other.manifest.sections():
+                return False
+            for option in self.manifest.options(s):
+                if(not other.manifest.has_option(s, option) or
+                   self.manifest.get(s, option) != other.manifest.get(s, option)):
+                    return False
         return True
-
+                
     # act like a configparser if asking for a non-existent method.
     def __getattr__(self, name):
         return getattr(self.manifest, name)
 
 
+class ConfigException(Exception):
+    """ Specifies exception is with a config """
+
+
 class Config(object):
     """
     Class to reconcile manifests
@@ -190,15 +209,16 @@ class Config(object):
     source_location = None
     config = {}
 
-    def __init__(self, source=None, target=None, namespace=None):
+    def __init__(self, source=None, target=None, namespace=None, lib=lib):
         """
         Takes in a source and target Manifest object, with a namespace
         to override if it is desired.
         """
+        self.lib = lib
         self.source = source
         self.target = target
         if self.target:
-            self.source_location = self.target.source() 
+            self.source_location = self.target.source()
         if self.source and not self.source_location:
             self.source.source()
         # store raws to use on write
@@ -214,103 +234,69 @@ class Config(object):
         if self.source:
             self.__load_configs(self.source)
 
-    def grab_inputs(self, manifest):
+    def setups(self):
+        """ Return a list of the features which need to be setup. """
+        if not self.target:
+            raise ConfigException("Update method requires a target manifest!")
+        return [s for s in self.target.formula_sections()
+                if self.target.run_phase(s, "setup")
+                if not self.source or not self.source.has_section(s)]
+
+    def updates(self):
+        """ Return a list of features which need to be updated. """
+        if not self.target:
+            raise ConfigException("Update method requires a target manifest!")
+        if not self.source:
+            raise ConfigException("Update method requires a source manifest!")
+        return [s for s in self.target.formula_sections()
+                if self.target.run_phase(s, "update")
+                if self.source.has_section(s)]
+
+    def removes(self):
+        """ Return a list of features which need to be destroyed. """
+        if not self.source:
+            raise ConfigException("Remove method requires a source manifest!")
+        return [s for s in self.source.formula_sections()
+                if self.source.run_phase(s, "remove")
+                if not self.target or not self.target.has_section(s)]
+
+    def deactivations(self):
+        """ Return a list of the features which need to be deactivated. """
+        if not self.source:
+            raise ConfigException("Deactivations method requires a source manifest!")
+        return [s for s in self.source.formula_sections()
+                if self.source.run_phase(s, "deactivate")]
+
+    def activations(self):
+        """ Return a list of the features which need to be deactivated. """
+        if not self.source:
+            raise ConfigException("Activations method requires a source manifest!")
+        return [s for s in self.source.formula_sections()
+                if self.source.run_phase(s, "activate")]
+
+    def grab_inputs(self, manifest=None):
         """
-        Look for any inputs not already asked accounted for in the manifest, and
+        Look for any inputs not already accounted for in the manifest, and
         query the user for them.
         """
+        if not manifest:
+            manifest = self.target if self.target else self.source
+        old_manifest = self.source if manifest is self.target else None
         if manifest:
             for s in manifest.sections():
                 if manifest.has_option(s, 'inputs'):
                     for param, attributes in \
                             self.__parse_input_string(manifest.get(s, 'inputs')):
-                        default = (attributes['default'] if 'default' in attributes else None)
-                        secret = (attributes['secret'] if 'secret' in attributes else False)
-                        self.get_config(param, default=default, secret=secret)
-
-    def setups(self):
-        """
-        Return a dictionary with all the features that need to be setup.
-        >>> c.setups()
-        [('myrc', {'target': {'formula': 'sprinter.formulas.template'}})]
-        """
-        new_sections = []
-        for s in self.target.formula_sections():
-            if not self.source or not self.source.has_section(s):
-                new_sections.append((s, {"target": dict(self.target.items(s))}))
-        return new_sections
-
-    def updates(self):
-        """
-        Return a dictionary with all the features that need to be
-        updated.
-
-        >>> c.updates()
-        [('maven', {'source': {'formula': 'sprinter.formulas.unpack', 'specific_version': '2.10'}, 'target': {'formula': 'sprinter.formulas.unpack', 'specific_version': '3.0.4'}}), ('ant', {'source': {'formula': 'sprinter.formulas.unpack', 'specific_version': '1.8.4'}, 'target': {'formula': 'sprinter.formulas.unpack', 'specific_version': '1.8.4'}})]
-
-        >>> config_old_only.updates()
-        Traceback (most recent call last):
-          File "<stdin>", line 1, in ?
-        ManifestError: Update method requires a target manifest!
-        """
-        if not self.target:
-            raise ManifestError("Update method requires a target manifest!")
-        different_sections = []
-        for s in self.target.formula_sections():
-            if self.source.has_section(s):
-                target_dict = dict(self.target.items(s))
-                source_dict = dict(self.source.items(s))
-                different_sections.append((s, {"source": source_dict,
-                                               "target": target_dict}))
-        return different_sections
-
-    def destroys(self):
-        """
-        Return a dictionary with all the features that need to be
-        destroyed.
-        >>> c.destroys()
-        [('mysql', {'source': {'brew': 'mysql', 'formula': 'sprinter.formulas.package', 'apt-get': 'libmysqlclient\\nlibmysqlclient-dev'}})]
-        """
-        missing_sections = []
-        for s in self.source.formula_sections():
-            if not self.target or not self.target.has_section(s):
-                missing_sections.append((s, {"source": dict(self.source.items(s))}))
-        return missing_sections
-
-    def deactivations(self):
-        """
-        Return a dictionary of activation formulas.
-
-        >>> c.deactivations()
-        [('maven', {'source': {'formula': 'sprinter.formulas.unpack', 'specific_version': '2.10'}}), ('ant', {'source': {'formula': 'sprinter.formulas.unpack', 'specific_version': '1.8.4'}}), ('mysql', {'source': {'brew': 'mysql', 'formula': 'sprinter.formulas.package', 'apt-get': 'libmysqlclient\\nlibmysqlclient-dev'}})]
-        """
-        deactivation_sections = []
-        for s in self.source.formula_sections():
-            deactivation_sections.append((s, {"source": dict(self.source.items(s))}))
-        return deactivation_sections
-
-    def activations(self):
-        """
-        Return a dictionary of activation formulas.
-
-        >>> c.activations()
-        [('maven', {'source': {'formula': 'sprinter.formulas.unpack', 'specific_version': '2.10'}}), ('ant', {'source': {'formula': 'sprinter.formulas.unpack', 'specific_version': '1.8.4'}}), ('mysql', {'source': {'brew': 'mysql', 'formula': 'sprinter.formulas.package', 'apt-get': 'libmysqlclient\\nlibmysqlclient-dev'}})]
-        """
-        activation_sections = []
-        for s in self.source.formula_sections():
-            activation_sections.append((s, {"source": dict(self.source.items(s))}))
-        return activation_sections
-
-    def reloads(self):
-        """
-        return reload dictionaries
-        >>> c.reloads()
-        [('maven', {'source': {'formula': 'sprinter.formulas.unpack', 'specific_version': '2.10'}}), ('ant', {'source': {'formula': 'sprinter.formulas.unpack', 'specific_version': '1.8.4'}}), ('mysql', {'source': {'brew': 'mysql', 'formula': 'sprinter.formulas.package', 'apt-get': 'libmysqlclient\\nlibmysqlclient-dev'}})]
-        """
-        reload_sections = []
-        for s in self.source.formula_sections():
-                reload_sections.append((s, {"source": dict(self.source.items(s))}))
-        return reload_sections
+                        if old_manifest and old_manifest.has_option('config', param):
+                            self.config[param] = old_manifest.get('config', param)
+                        else:
+                            default = (attributes['default'] if 'default' in attributes else None)
+                            secret = (attributes['secret'] if 'secret' in attributes else False)
+                            self.get_config(param, default=default, secret=secret)
+            if self.target:
+                self.set_additional_context('target')
+            if self.source:
+                self.set_additional_context('source')
 
     def write(self, file_handle):
         """
@@ -332,28 +318,26 @@ class Config(object):
         """
         grabs a config from the user space; if it doesn't exist, it will prompt for it.
         """
-        prompt = "please enter your %s" % param_name
-        if default:
-            prompt += " (default %s)" % default
         if param_name not in self.config:
-            self.config[param_name] = lib.prompt("please enter your %s" % param_name, default=default, secret=secret)
+            self.config[param_name] = self.lib.prompt("please enter your %s" % param_name,
+                                                      default=default,
+                                                      secret=secret)
         if secret:
             self.temporary_sections.append(param_name)
         return self.config[param_name]
 
-    def get_context_dict(self, kind='target'):
+    def set_additional_context(self, manifest_type='target', additional_context={}):
         """
         return a context dict of the desired state
         """
-        context_dict = {}
-        manifest = self.target if kind == 'target' else self.source
-        if manifest:
-            for s in manifest.formula_sections():
-                for k, v in manifest.items(s):
-                    context_dict["%s:%s" % (s, k)] = v
+        if not hasattr(self, manifest_type):
+            raise ConfigException("manifest_type '%s' doesn't exist!" % manifest_type)
+        manifest = getattr(self, manifest_type)
+        context_dict = manifest.additional_context_variables
         for k, v in self.config.items():
                 context_dict["config:%s" % k] = v
-        return context_dict
+        manifest.additional_context_variables = dict(context_dict.items() +
+                                                     additional_context.items())
 
     def __load_configs(self, config):
         if config.has_section('config'):
@@ -377,29 +361,6 @@ class Config(object):
             namespace = s
         return namespace
 
-    def __update_needed(self, source_dict, target_dict):
-        """
-        checks if an update is needed if there is a diff between the items provided
-
-        >>> c._Config__update_needed({"a": "b", "c": "d"}, {"a": "b", "c": "e"})
-        True
-
-        >>> c._Config__update_needed({"a": "b"}, {"a": "b", "c": "e"})
-        True
-
-        >>> c._Config__update_needed({"a": "b", "c": "d"}, {"a": "b"})
-        True
-
-        >>> c._Config__update_needed({"a": "b", "c": "d", "e": "f"}, {"a": "b", "c": "d", "e": "f"})
-        False
-        """
-        if len(source_dict) != len(target_dict):
-            return True
-        for k in source_dict:
-            if k not in target_dict or source_dict[k] != target_dict[k]:
-                return True
-        return False
-
     def __parse_input_string(self, input_string):
         """
         parse an attribute in a given input string format:
@@ -410,31 +371,12 @@ class Config(object):
                  username
                  password?
                  main_branch==comp_main
-
-        >>> c._Config__parse_input_string(test_input_string)
-        [('gitroot', {'default': '~/workspace'}), ('username', {}), ('password', {'secret': True}), ('main_branch', {'default': 'comp_main'})]
         """
         raw_params = input_string.split('\n')
         return [self.__parse_param_line(rp) for rp in raw_params if len(rp.strip(' \t')) > 0]
 
     def __parse_param_line(self, line):
-        """
-        Parse a single param line.
-
-        >>> c._Config__parse_param_line("username")
-        ('username', {})
-
-        >>> c._Config__parse_param_line("password?")
-        ('password', {'secret': True})
-
-        >>> c._Config__parse_param_line("main_branch==comp_main")
-        ('main_branch', {'default': 'comp_main'})
-
-        >>> c._Config__parse_param_line("main_branch?==comp_main")
-        ('main_branch', {'default': 'comp_main', 'secret': True})
-
-        >>> c._Config__parse_param_line("")
-        """
+        """ Parse a single param line. """
         value = line.strip('\n \t')
         if len(value) > 0:
             attribute_dict = {}
@@ -447,13 +389,17 @@ class Config(object):
             return (value, attribute_dict)
         return None
 
-if __name__ == '__main__':
-    import doctest
-    old_manifest = Manifest(StringIO(test_old_version))
-    new_manifest = Manifest(StringIO(test_new_version))
-    config = Config(source=old_manifest, target=new_manifest)
-    config_new_only = Config(target=new_manifest)
-    config_old_only = Config(source=old_manifest)
-    doctest.testmod(extraglobs={'c': config,
-                                'config_new_only': config_new_only,
-                                'config_old_only': config_old_only})
+    def context(self, manifest_type=None):
+        """ get the context dictionary desired """
+        manifest = None
+        if manifest_type is None:
+            manifest = self.target if self.target else self.source
+        else:
+            manifest = getattr(self, manifest_type)
+        return manifest.get_context_dict()
+
+    def set_source(self, source_manifest):
+        """ Set the source manifest """
+        self.source = source_manifest
+        self.set_additional_context('source')
+        return source_manifest
